<?php
/**
 +-----------------------------------------------------------------------+
 | common.php                                                            |
 |                                                                       |
 | This file is part of the import/export contacts script                |
 |                                                                       |
 | Copyright (C) 2018, Cliss XXI                                         |
 |                                                                       |
 | Licensed under the GNU General Public License version 3 or            |
 | any later version with exceptions for skins & plugins.                |
 | See the LICENSE file for a full license statement.                    |
 |                                                                       |
 +-----------------------------------------------------------------------+
 | Author: Jerome Lebleu <contact@cliss21.com>                           |
 +-----------------------------------------------------------------------+
 */

define('INSTALL_PATH', realpath(__DIR__ . '/..') . '/');

require_once INSTALL_PATH . 'program/include/clisetup.php';

// Logging facilities
// ------------------

$LOG_LEVELS = array(
    'DEBUG' => 0,
    'INFO' => 1,
    'WARNING' => 2,
    'ERROR' => 3,
);

$LOG_LEVEL = $LOG_LEVELS['WARNING'];

/**
 * Set the log level to use in the script.
 *
 * @param string $level Log level
 */
function set_log_level($level)
{
    global $LOG_LEVEL, $LOG_LEVELS;

    if (isset($LOG_LEVELS[$level])) {
        $LOG_LEVEL = $LOG_LEVELS[$level];
    }
}

/**
 * Check whether a level is logged.
 *
 * @param string $level Log level
 *
 * @return bool
 */
function is_logged_level($level)
{
    global $LOG_LEVEL, $LOG_LEVELS;

    if (!isset($LOG_LEVELS[$level])) {
        $level = 'DEBUG';
    }
    if ($LOG_LEVELS[$level] >= $LOG_LEVEL) {
        return true;
    }

    return false;
}

/**
 * Log a message at a given level.
 *
 * @param string $level   Log level
 * @param string $message Message
 */
function log_msg($level, $message)
{
    if (is_logged_level($level)) {
        echo "[$level] $message\n";
    }
}

/**
 * Display error and terminate script execution.
 *
 * @param string $message Message
 * @param int    $status  Optional exit status
 */
function raise_err($message, $status = 1)
{
    fputs(STDERR, "Error: $message\n");
    exit($status);
}

// Import and export helpers
// -------------------------

/**
 * Copy contact record properties into a vcard object.
 *
 * @see program/steps/addressbook/export.inc
 */
function prepare_for_export(&$record, $source = null)
{
    $groups = $source && $source->groups && $source->export_groups ? $source->get_record_groups($record['ID']) : null;
    $fieldmap = $source ? $source->vcard_map : null;

    if (empty($record['vcard'])) {
        $vcard = new rcube_vcard($record['vcard'], RCUBE_CHARSET, false, $fieldmap);
        $vcard->reset();

        foreach ($record as $key => $values) {
            list($field, $section) = explode(':', $key);
            // avoid unwanted casting of DateTime objects to an array
            // (same as in rcube_contacts::convert_save_data())
            if (is_object($values) && is_a($values, 'DateTime')) {
                $values = array($values);
            }

            foreach ((array) $values as $value) {
                if (is_array($value) || is_a($value, 'DateTime') || @strlen($value)) {
                    $vcard->set($field, $value, strtoupper($section));
                }
            }
        }

        // append group names
        if ($groups) {
            $vcard->set('groups', join(',', $groups), null);
        }

        $record['vcard'] = $vcard->export();
    }
    // patch categories to alread existing vcard block
    elseif ($record['vcard']) {
        $vcard = new rcube_vcard($record['vcard'], RCUBE_CHARSET, false, $fieldmap);

        // unset CATEGORIES entry, it might be not up-to-date (#1490277)
        $vcard->set('groups', null);
        $record['vcard'] = $vcard->export();

        if (!empty($groups)) {
            $vgroups = 'CATEGORIES:' . rcube_vcard::vcard_quote($groups, ',');
            $record['vcard'] = str_replace('END:VCARD', $vgroups . rcube_vcard::$eol . 'END:VCARD', $record['vcard']);
        }
    }
}

/**
 * Returns the matching group id. If group doesn't exist, it'll be created if allowed.
 *
 * @see program/steps/addressbook/import.inc
 */
function import_group_id($group_name, $contacts, $create, &$import_groups)
{
    $group_id = 0;
    foreach ($import_groups as $group) {
        if (strtolower($group['name']) == strtolower($group_name)) {
            $group_id = $group['ID'];
            break;
        }
    }

    // create a new group
    if (!$group_id && $create) {
        $new_group = $contacts->create_group($group_name);
        if (!$new_group['ID']) {
            $new_group['ID'] = $new_group['id'];
        }
        $import_groups[] = $new_group;
        $group_id = $new_group['ID'];
    }

    return $group_id;
}

// Main class
// ----------

class rcmail_contacts_utils
{
    private $host = null;
    private $username = null;

    private $db = null;
    private $user_id = 0;
    private $contacts = null;

    /**
     * Object constructor.
     *
     * @param string $user User name
     * @param string $host Optional user host
     */
    public function __construct($user, $host = null, $collected_contacts = false)
    {
        $this->db = rcmail_utils::db();

        $this->host = $host;
        $this->username = $user;

        $this->user_id = $this->get_user_id($this->username, $this->host);
	if ($collected_contacts) {
	  log_msg('INFO', 'working on collected_contacts - if script stop here you probably need to install automatic_addressbook plugin');
	  require_once INSTALL_PATH . 'plugins/automatic_addressbook/automatic_addressbook_backend.php';
	  $this->contacts = new automatic_addressbook_backend($this->db, $this->user_id);
	} else {
	  log_msg('INFO', 'working on contacts');
	  $this->contacts = new rcube_contacts($this->db, $this->user_id);
	}
    }

    /**
     * Export the user's address book as vCard file.
     *
     * @param resource $out Output file handler
     */
    public function export($out)
    {
        // get contacts for this user
        $this->contacts->set_page(1);
        $this->contacts->set_pagesize(99999);
        $result = $this->contacts->list_records(null, 0, true);

        if (0 == $result->count) {
            raise_err("No address book found for user {$this->username}");
        }

        while ($result && ($row = $result->next())) {
            prepare_for_export($row, $this->contacts);

            // fix folding and end-of-line chars
            $row['vcard'] = preg_replace('/\r|\n\s+/', '', $row['vcard']);
            $row['vcard'] = preg_replace('/\n/', rcube_vcard::$eol, $row['vcard']);
            fwrite($out, rcube_vcard::rfc2425_fold($row['vcard']) . rcube_vcard::$eol);
        }

        log_msg('INFO', "{$result->count} contacts have been exported for {$this->username}");
    }

    /**
     * Import contacts from a vCard file to user's address book.
     *
     * @param resource $in Input file handler
     */
    public function import($in, $replace = false, $confirm = true)
    {
        $with_groups = (!$this->contacts->groups) ? false : true;

        $file_content = stream_get_contents($in);

        log_msg('DEBUG', "Content of the input file to import: << EOF\n${file_content}EOF");

        $vcard_o = new rcube_vcard();
        $vcard_o->extend_fieldmap($this->contacts->vcard_map);
        $vcards = $vcard_o->import($file_content);

        if (empty($vcards)) {
            raise_err('No contact found in the input');
        }

        $stats = new stdClass();
        $stats->names = array();
        $stats->skipped_names = array();
        $stats->count = count($vcards);
        $stats->inserted = $stats->skipped = $stats->invalid = $stats->errors = 0;

        if ($confirm) {
            echo "{$stats->count} contacts will be imported. Do you want to continue? (y/N) ";
            $input = trim(fgets(STDIN));

            if ('y' != strtolower($input)) {
                return false;
            }
        } else {
            log_msg('DEBUG', "{$stats->count} contacts founds in the input file");
        }

        if ($replace) {
            $this->contacts->delete_all($this->contacts->groups && $with_groups < 2);
        }

        if ($with_groups) {
            $import_groups = $this->contacts->list_groups();
        }

        foreach ($vcards as $vcard) {
            $a_record = $vcard->get_assoc();

            // Generate contact's display name (must be before validation), the same we do in save.inc
            if (empty($a_record['name'])) {
                $a_record['name'] = rcube_addressbook::compose_display_name($a_record, true);
                // Reset it if equals to email address (from compose_display_name())
                if ($a_record['name'] == $a_record['email'][0]) {
                    $a_record['name'] = '';
                }
            }

            // skip invalid (incomplete) entries
            if (!$this->contacts->validate($a_record, true)) {
                ++$stats->invalid;
                continue;
            }

            // We're using UTF8 internally
            $email = $vcard->email[0];
            $email = rcube_utils::idn_to_utf8($email);

            if (!$replace) {
                $existing = null;
                // compare e-mail address
                if ($email) {
                    $existing = $this->contacts->search('email', $email, 1, false);
                }
                // compare display name if email not found
                if ((!$existing || !$existing->count) && $vcard->displayname) {
                    $existing = $this->contacts->search('name', $vcard->displayname, 1, false);
                }
                if ($existing && $existing->count) {
                    ++$stats->skipped;
                    $stats->skipped_names[] = $vcard->displayname ?: $email;
                    continue;
                }
            }

            $a_record['vcard'] = $vcard->export();

            // insert record
            $success = $this->contacts->insert($a_record);

            if ($success) {
                // assign groups for this contact (if enabled)
                if ($with_groups && !empty($a_record['groups'])) {
                    foreach (explode(',', $a_record['groups'][0]) as $group_name) {
                        if ($group_id = import_group_id($group_name, $this->contacts, 1 == $with_groups, $import_groups)) {
                            $this->contacts->add_to_group($group_id, $success);
                        }
                    }
                }

                ++$stats->inserted;
                $stats->names[] = $a_record['name'] ?: $email;
            } else {
                ++$stats->errors;
            }
        }

        echo "Importation results:\n";
        echo " - inserted: {$stats->inserted}\n";
        echo " - skipped: {$stats->skipped}\n";
        echo " - invalid: {$stats->invalid}\n";
        echo " - errors: {$stats->errors}\n";

        if (is_logged_level('INFO')) {
            log_msg('INFO', 'Names of imported contacts: ' . implode(', ', $stats->names));
            if (!empty($stats->skipped_names)) {
                log_msg('INFO', 'Names of skipped contacts: ' . implode(', ', $stats->skipped_names));
            }
        }
    }

    /**
     * Find the user id matching the given name and host.
     *
     * @param string $user User name
     * @param string $host Optional user host
     *
     * @return int User id
     */
    private function get_user_id($user, $host = null)
    {
        // construct SQL query
        $sql = 'SELECT user_id, mail_host, last_login ' .
            'FROM ' . $this->db->table_name('users', true) . ' WHERE `username` = ?';
        $sql_params = array($user);
        if ($host) {
            $sql .= ' AND `mail_host` = ?';
            $sql_params[] = $host;
        }

        log_msg('DEBUG', "Search user ${user} with SQL query: ${sql} / " .
            print_r($sql_params, 1));

        $sql_result = $this->db->query($sql, $sql_params);

        if ($sql_result) {
            if (0 == $this->db->num_rows($sql_result)) {
                raise_err("User ${user} not found");
            } elseif ($this->db->num_rows($sql_result) > 1) {
                // output users details
                if (is_logged_level('INFO')) {
                    $log_msg = $this->db->num_rows($sql_result) .
                        " users found for ${user}:";
                    while ($sql_arr = $this->db->fetch_assoc($sql_result)) {
                        $log_msg .= "\n - id: " . $sql_arr['user_id'] .
                            ' / mail_host: ' . $sql_arr['mail_host'] .
                            ' / last_login: ' . $sql_arr['last_login'];
                    }
                    log_msg('INFO', $log_msg);
                }

                raise_err("More thant one user found for username ${user}, you need to specify the host");
            } else {
                $sql_arr = $this->db->fetch_assoc($sql_result);
                $user_id = $sql_arr['user_id'];

                log_msg('DEBUG', "Found user ${user} with id ${user_id}");

                return $user_id;
            }
        }

        raise_err("Unable to query the user ${user}!");
    }
}
